package com.thebrokenrail.energonrelics.energy.core;

import com.thebrokenrail.energonrelics.component.NetworkComponent;
import com.thebrokenrail.energonrelics.energy.core.util.Action;
import com.thebrokenrail.energonrelics.energy.core.util.EnergyProvider;
import net.minecraft.block.BlockState;
import net.minecraft.block.entity.BlockEntityType;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.IntArrayTag;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.Tickable;
import net.minecraft.util.math.Vec3d;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public abstract class EnergyReceiverBlockEntity extends EnergyProviderBlockEntity implements Tickable {
    public EnergyReceiverBlockEntity(BlockEntityType<?> type) {
        super(type);
    }

    private final List<Action.PropagatedAction> pendingPropagatedActions = new ArrayList<>();
    private final List<Action.PropagatedAction> pendingActions = new ArrayList<>();
    private final List<EnergyProvider> providers = new ArrayList<>();
    private final List<Integer> networks = new ArrayList<>();

    private long totalCost = 0;

    protected void propagateAction(Action.PropagatedAction action) {
        totalCost = totalCost + action.amountOwed();
        pendingActions.add(action);
    }

    protected void addAction(Action action) {
        propagateAction(new Action.PropagatedActionImpl(action, getWorld(), getPos(), getCachedState().getBlock()));
    }

    @Override
    public void serverTick() {
        super.serverTick();

        ServerWorld world = (ServerWorld) getWorld();
        assert world != null;
        NetworkComponent component = NetworkComponent.getInstance(world);

        providers.clear();
        for (int network : networks) {
            List<EnergyProvider> potentialProviders = component.getSource(world, network);
            for (EnergyProvider provider : potentialProviders) {
                if (provider.getRegistryKey().equals(getWorld().getRegistryKey())) {
                    Vec3d pos = new Vec3d(getPos().getX() + 0.5d, getPos().getY() + 0.5d, getPos().getZ() + 0.5d);
                    if (provider.isWithinDistance(pos)) {
                        providers.add(provider);
                    }
                }
            }
        }

        totalCost = 0;
        pendingActions.clear();
        tickEnergy();

        for (Action.PropagatedAction action : pendingPropagatedActions) {
            propagateAction(action);
        }
        pendingPropagatedActions.clear();

        for (Action.PropagatedAction action : pendingActions) {
            if (providers.size() > 0) {
                action.expandPayments(providers.size());
                for (EnergyProvider provider : providers) {
                    provider.addPropagatedAction(action);
                }
            } else {
                action.pay(0);
            }
        }
    }

    public long getTotalCost() {
        return totalCost;
    }

    @Override
    protected void handlePropagatedAction(Action.PropagatedAction action) {
        super.handlePropagatedAction(action);
        // Propagate Action To Energy Providers
        if (isEnergyProviderActive()) {
            pendingPropagatedActions.add(action);
        } else {
            action.pay(0);
        }
    }

    @Override
    public CompoundTag toTag(CompoundTag tag) {
        super.toTag(tag);
        tag.put("Networks", new IntArrayTag(networks));
        return tag;
    }

    @Override
    public void fromTag(BlockState state, CompoundTag tag) {
        super.fromTag(state, tag);
        int[] networksArray = tag.getIntArray("Networks");
        networks.clear();
        for (int network : networksArray) {
            networks.add(network);
        }
    }

    protected boolean isEnergyProviderActive() {
        throw new UnsupportedOperationException();
    }

    protected abstract void tickEnergy();

    public void toggle(int network) {
        if (contains(network)) {
            networks.removeAll(Collections.singletonList(network));
        } else {
            networks.add(network);
        }
        markDirty();
    }

    public boolean contains(int network) {
        return networks.contains(network);
    }
}
